@@ -13,7 +13,7 @@ INTERPOLATION_MODE_NEAREST = 0
 INTERPOLATION_MODE_LINEAR = 1
 
 # Advanced settings
-COM_RADIUS = 2
+COM_RADIUS = 4
 COS_FILTER_POWER = 0.5  # input images are multiplied by a cosine mask, with a value 1.0 in the center of the image and 0.0 at the edges; e.g. for an N x N sized image, the mask is (cos((px_x - N//2) * pi / N) * cos((px_y - N//2) * pi / N)) ** COS_FILTER_POWER, with px_x/y the pixel coordinate. Raising the filter to a low power, e.g. 0.33, makes it affect the central region of the image less.
 GLFW_CONTEXT_VERSION_MAJOR = 3  # user can change the GLFW/GL version prior to calling pyGPUreg.init() but code was not tested with any other version than 3.3
 GLFW_CONTEXT_VERSION_MINOR = 3  # user can change the GLFW/GL version prior to calling pyGPUreg.init() but code was not tested with any other version than 3.3
@@ -47,7 +47,9 @@ texture_rg_T: Texture
 texture_rg_I: Texture
 texture_rg_buffer: Texture
 
-def init(create_window=True, image_size=None):
+
+
+def init():
     """
     Initialize pyGPUreg.
     :param create_window: bool (default True). When True, pyGPUreg creates a glfw window context for OpenGL. When using pyGPUreg within a project that already has an OpenGL context, no window needs to be created.
@@ -55,18 +57,17 @@ def init(create_window=True, image_size=None):
     """
     global window, cs_butterfly, cs_cosft, cs_fft, cs_fft_pi, cs_multiply, cs_resample, cs_fft_single, cs_fft_pi_single, cs_multiply_single, cs_copy_r_to_rg, cs_cosft_single
 
-    if create_window:
-        if not glfw.init():
-            raise Exception("Could not initialize GLFW")
+    if not glfw.init():
+        raise Exception("Could not initialize GLFW")
 
-        # create a hidden glfw window and set OpenGL version
-        glfw.window_hint(glfw.VISIBLE, glfw.FALSE)
-        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, GLFW_CONTEXT_VERSION_MAJOR)
-        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, GLFW_CONTEXT_VERSION_MINOR)
-        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
-        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, GL_TRUE)
-        window = glfw.create_window(2, 2, "pyGPUreg hidden window", None, None)
-        glfw.make_context_current(window)
+    # create a hidden glfw window and set OpenGL version
+    glfw.window_hint(glfw.VISIBLE, glfw.FALSE)
+    glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, GLFW_CONTEXT_VERSION_MAJOR)
+    glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, GLFW_CONTEXT_VERSION_MINOR)
+    glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
+    glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, GL_TRUE)
+    window = glfw.create_window(2, 2, "pyGPUreg hidden window", None, None)
+    glfw.make_context_current(window)
 
     # compile shaders
     shader_dir = os.path.join(os.path.dirname(__file__))+"\\shaders\\"
@@ -82,9 +83,6 @@ def init(create_window=True, image_size=None):
     cs_multiply_single = Shader(shader_dir + "fft_phase_correlation_single.glsl")
     cs_copy_r_to_rg = Shader(shader_dir + "copy_r_to_rg.glsl")
     cs_cosft_single = Shader(shader_dir + "cos_filter_r.glsl")
-    # create textures with set_image_size
-    if image_size is not None:
-        set_image_size(image_size)
 
 
 def set_image_size(size):
@@ -155,6 +153,7 @@ def set_image_size(size):
 
     cs_butterfly.unbind()
 
+
 def detect_subpixel_maximum(pcorr, mode):
     shift = np.zeros(2)
     if mode == SUBPIXEL_MODE_NONE:
@@ -170,13 +169,15 @@ def detect_subpixel_maximum(pcorr, mode):
                 shift[0] += x * pc_val
                 shift[1] += y * pc_val
                 mass += pc_val
-        shift /= mass
+        if mass != 0.0:
+            shift /= mass
+        else:
+            shift = (image_size / 2, image_size / 2)
 
     x_shift = image_size / 2 - shift[0]
     y_shift = image_size / 2 - shift[1]
-    if log_image_size % 2 == 0:  # when logN is even, change the sign of the shift.
-        x_shift = -x_shift
-        y_shift = -y_shift
+    x_shift = -x_shift
+    y_shift = -y_shift
     return x_shift, y_shift
 
 def sample_texture_with_shift(texture, shift, edge_mode=EDGE_MODE_ZERO, interpolation_mode=INTERPOLATION_MODE_LINEAR):
@@ -190,7 +191,7 @@ def sample_texture_with_shift(texture, shift, edge_mode=EDGE_MODE_ZERO, interpol
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
     elif edge_mode == EDGE_MODE_REFLECT:
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
-        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
+        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT)
     if interpolation_mode == INTERPOLATION_MODE_NEAREST:
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
@@ -198,13 +199,17 @@ def sample_texture_with_shift(texture, shift, edge_mode=EDGE_MODE_ZERO, interpol
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
 
-    texture_resample_b.update(None, width=image_size, height=image_size)  # empty the buffer texture into which the original image is copied
+    width = image_size
+    height = image_size
+
+    texture_resample_b.update(None, width=height, height=width)  # empty the buffer texture into which the original image is copied
     cs_resample.bind()  # bind compute shader, upload uniforms, bind textures, dispatch
     cs_resample.uniform1f("dx", float(shift[0]))
     cs_resample.uniform1f("dy", float(shift[1]))
-    cs_resample.uniform1i("N", image_size)
+    cs_resample.uniform1i("W", width)
+    cs_resample.uniform1i("H", height)
     cs_resample.uniform1i("edge_mode", edge_mode)
-    texture_resample_a.bind(0)
+    texture.bind(0)
     glBindImageTexture(1, texture_resample_b.renderer_id, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32F)
     glDispatchCompute(*compute_space_size)
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
@@ -226,6 +231,9 @@ def sample_image_with_shift(image, shift, edge_mode=EDGE_MODE_ZERO, interpolatio
     :param interpolation_mode: one of: pypgureg.INTERPOLATION_MODE_LINEAR (default), .INTERPOLATION_MODE_NEAREST. Note that combining nearest neighbour interpolation with a shift less than one pixel causes the output to be identical to the input image.
     :return: numpy array of pixel values in the resampled image.
     """
+    if window is not None:
+        previous_context = glfw.get_current_context()
+        glfw.make_context_current(window)
     # next up: re-sampling the image and apply the detected shift.
     texture_resample_a.update(image)  # upload the image to be resampled to the gpu
     texture_resample_a.bind()
@@ -235,7 +243,7 @@ def sample_image_with_shift(image, shift, edge_mode=EDGE_MODE_ZERO, interpolatio
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
     elif edge_mode == EDGE_MODE_REFLECT:
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
-        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
+        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT)
     if interpolation_mode == INTERPOLATION_MODE_NEAREST:
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
@@ -243,15 +251,19 @@ def sample_image_with_shift(image, shift, edge_mode=EDGE_MODE_ZERO, interpolatio
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
         glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
 
-    texture_resample_b.update(None, width=image_size, height=image_size) # empty the buffer texture into which the original image is copied
+    width = image.shape[1]
+    height = image.shape[0]
+
+    texture_resample_b.update(None, width=width, height=height)  # empty the buffer texture into which the original image is copied
     cs_resample.bind()  # bind compute shader, upload uniforms, bind textures, dispatch
     cs_resample.uniform1f("dx", float(shift[0]))
     cs_resample.uniform1f("dy", float(shift[1]))
-    cs_resample.uniform1i("N", image_size)
+    cs_resample.uniform1i("W", width)
+    cs_resample.uniform1i("H", height)
     cs_resample.uniform1i("edge_mode", edge_mode)
     texture_resample_a.bind(0)
     glBindImageTexture(1, texture_resample_b.renderer_id, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32F)
-    glDispatchCompute(*compute_space_size)
+    glDispatchCompute(1 + width // 16, 1 + height // 16, 1)
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
     cs_resample.unbind()
 
@@ -260,6 +272,10 @@ def sample_image_with_shift(image, shift, edge_mode=EDGE_MODE_ZERO, interpolatio
     resampled = glGetTexImage(GL_TEXTURE_2D, 0, GL_RED, GL_FLOAT)
     glMemoryBarrier(GL_TEXTURE_FETCH_BARRIER_BIT)
 
+    resampled = np.reshape(resampled, (height, width))
+
+    if window is not None:
+        glfw.make_context_current(previous_context)
     return resampled
 
 
@@ -323,6 +339,9 @@ def bind_and_launch_cos_filter_compute_shader():
 
 
 def gpu_fft(image, image2=None):
+    if window is not None:
+        previous_context = glfw.get_current_context()
+        glfw.make_context_current(window)
     """
     Compute a FFT on the GPU for one of two input images. The data is sent to the GPU in the form of an RGBA, float32
     texture. Since one FFT requires only two colour channels (one for the real and one for the imaginary part), it is
@@ -356,6 +375,8 @@ def gpu_fft(image, image2=None):
 
     texture_data.bind()
     fourier_transforms = glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT)
+    if window is not None:
+        glfw.make_context_current(previous_context)
     if image2 is not None:
         return fourier_transforms[:, :, 0] + 1j * fourier_transforms[:, :, 1], fourier_transforms[:, :, 2] + 1j * fourier_transforms[:, :, 3]
     else:
@@ -363,6 +384,9 @@ def gpu_fft(image, image2=None):
 
 
 def register(template_image, moved_image, apply_shift=True, edge_mode=EDGE_MODE_ZERO, subpixel_mode=SUBPIXEL_MODE_COM, interpolation_mode=INTERPOLATION_MODE_LINEAR):
+    if window is not None:
+        previous_context = glfw.get_current_context()
+        glfw.make_context_current(window)
     """
     Register moved_image onto template_image by computing the phase correlation of a template and moved image in order to detect the shift between template and moved image. Optionally also resamples the moved image to undo the shift.
     :param template_image: 2d numpy array of pixel data for the template image.
@@ -384,8 +408,6 @@ def register(template_image, moved_image, apply_shift=True, edge_mode=EDGE_MODE_
         raise Exception(
             f"Image size is {template_image.shape} but should be ({image_size}, {image_size}). Call pyGPUreg.set_image_size() to change the expected image size")
 
-    import time
-    timer = time.time_ns()
 
     data = np.zeros((image_size, image_size, 4))
     data[:, :, 0] = template_image
@@ -400,7 +422,6 @@ def register(template_image, moved_image, apply_shift=True, edge_mode=EDGE_MODE_
     # forward FFTs
     bind_and_launch_fft_compute_shaders()
 
-
     # use the cs_multiply compute shader to calculate the product of the Fourier transforms
     cs_multiply.bind()
     glBindImageTexture(0, texture_data.renderer_id, 0, GL_FALSE, 0, GL_READ_WRITE, GL_RGBA32F)
@@ -408,7 +429,6 @@ def register(template_image, moved_image, apply_shift=True, edge_mode=EDGE_MODE_
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
     cs_multiply.unbind()
 
-
     # forward FFTs again to get cross correlation (without amplitude scaling.)
     bind_and_launch_fft_compute_shaders(do_inversion_permutation=False)
     glMemoryBarrier(GL_TEXTURE_UPDATE_BARRIER_BIT)
@@ -416,13 +436,17 @@ def register(template_image, moved_image, apply_shift=True, edge_mode=EDGE_MODE_
     # get phase correlation image and find maximum
     texture_data.bind()
     pcorr = np.fft.fftshift(glGetTexImage(GL_TEXTURE_2D, 0, GL_RED, GL_FLOAT), axes=(0, 1))
-    #pcorr[image_size//2, image_size//2] = 0
+    m = image_size//2
+    pcorr[m, m] = (pcorr[m + 1, m] + pcorr[m - 1, m] + pcorr[m, m + 1] + pcorr[m, m - 1]) / 4
     dx, dy = detect_subpixel_maximum(pcorr, mode=subpixel_mode)
     if not apply_shift:
+        if window is not None:
+            glfw.make_context_current(previous_context)
         return dx, dy
 
-
-    resampled = sample_image_with_shift(moved_image, (dx, dy), edge_mode, interpolation_mode)
+    resampled = sample_image_with_shift(moved_image, (dy, dx), edge_mode, interpolation_mode)
+    if window is not None:
+        glfw.make_context_current(previous_context)
     return resampled, (dx, dy)
 
 
@@ -451,7 +475,7 @@ def bind_and_launch_fft_single_compute_shaders(ft_texture, do_inversion_permutat
         pingpong += 1
 
     cs_fft_single.unbind()
-
+    glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
     if not do_inversion_permutation:
         return
 
@@ -464,28 +488,21 @@ def bind_and_launch_fft_single_compute_shaders(ft_texture, do_inversion_permutat
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
     cs_fft_pi_single.unbind()
 
-    # FFT of the data that was initially in ft_texture r channel is now in ft_texture
-
-def _debug_show_img(texture, title):
-    texture.bind()
-    img = glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT)
-    plt.subplot(1, 2, 1)
-    plt.imshow(img[:, :, 0])
-    plt.title(title)
-    plt.subplot(1,2,2)
-    plt.imshow(img[:, :, 1])
-    plt.show()
-
-
 
 def set_template(template_image):
+    if window is not None:
+        previous_context = glfw.get_current_context()
+        glfw.make_context_current(window)
     s = template_image.shape
-    if s[0] != image_size or s[1] != image_size:
-        raise Exception(f"Template image must be square and size {image_size}, or change size by calling pyGPUfit.set_image_size()")
+    if s[0] != s[1]:
+        raise Exception(f"pyGPUreg: input image must be square and 2D. Current image has shape: {s}")
+    if s[0] != image_size:
+        set_image_size(s[0])
 
     # upload to GPU
     cs_copy_r_to_rg.bind()
     texture_r_t.update(template_image)
+    glMemoryBarrier(GL_TEXTURE_UPDATE_BARRIER_BIT)
     glBindImageTexture(0, texture_r_t.renderer_id, 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32F)
     glBindImageTexture(1, texture_rg_T.renderer_id, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RG32F)
     glDispatchCompute(*compute_space_size)
@@ -502,13 +519,19 @@ def set_template(template_image):
 
     # compute FFT
     bind_and_launch_fft_single_compute_shaders(texture_rg_T)
+    if window is not None:
+        glfw.make_context_current(previous_context)
+
 
 
 def register_to_template(image, apply_shift=True, edge_mode=EDGE_MODE_ZERO, subpixel_mode=SUBPIXEL_MODE_COM, interpolation_mode=INTERPOLATION_MODE_LINEAR):
+    if window is not None:
+        previous_context = glfw.get_current_context()
+        glfw.make_context_current(window)
     s = image.shape
     if s[0] != image_size or s[1] != image_size:
         raise Exception(
-            f"Template image must be square and size {image_size}, or change size by calling pyGPUfit.set_image_size()")
+            f"Image must be square and size {image_size}, or change size by calling pyGPUfit.set_image_size()")
 
     # upload to GPU
     cs_copy_r_to_rg.bind()
@@ -519,7 +542,6 @@ def register_to_template(image, apply_shift=True, edge_mode=EDGE_MODE_ZERO, subp
     cs_copy_r_to_rg.unbind()
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
 
-
     # apply cosine filter
     cs_cosft_single.bind()
     glBindImageTexture(0, texture_cos_mask.renderer_id, 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32F)
@@ -527,7 +549,6 @@ def register_to_template(image, apply_shift=True, edge_mode=EDGE_MODE_ZERO, subp
     glDispatchCompute(*compute_space_size)
     glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
     cs_cosft_single.unbind()
-
     # compute FFT
     bind_and_launch_fft_single_compute_shaders(texture_rg_I)
     # compute phase correlation
@@ -544,10 +565,16 @@ def register_to_template(image, apply_shift=True, edge_mode=EDGE_MODE_ZERO, subp
     # get phase correlation image and find maximum
     texture_rg_I.bind()
     pcorr = np.fft.fftshift(glGetTexImage(GL_TEXTURE_2D, 0, GL_RED, GL_FLOAT), axes=(0, 1))
-    #pcorr[image_size//2, image_size//2] = 0
+    m = image_size//2
+    pcorr[m, m] = (pcorr[m+1, m] + pcorr[m-1, m] + pcorr[m, m+1] + pcorr[m, m-1]) / 4
     dx, dy = detect_subpixel_maximum(pcorr, mode=subpixel_mode)
     if not apply_shift:
+        if window is not None:
+            glfw.make_context_current(previous_context)
         return dx, dy
 
     resampled = sample_texture_with_shift(texture_r_i, (dx, dy), edge_mode, interpolation_mode)
-    return resampled, (dx, dy)
\ No newline at end of file
+    if window is not None:
+        glfw.make_context_current(previous_context)
+    return resampled, (dx, dy)
+
